
//Title:        Emulador para o Kit MC8051
//Version:
//Copyright:    Copyright (c) 1997
//Author:       Hermes F. Reis Neto / Jan L. de Lara / Kleber W. Tolezani
//Company:      Escola Politcnica da Universidade de So Paulo
//Description:


package MC8051;

//import java.awt.Frame;
//import java.awt.BorderLayout;
//import java.awt.Dimension;
//import java.awt.ToolKit;
import java.awt.*;

/*
** Numerao das teclas (0 a 15) em funo dos bits de P1 (0 = menos sig.):
    0  - bits 0 e 4
    1  - bits 0 e 5
    2  - bits 0 e 6
    3  - bits 0 e 7
    4  - bits 1 e 4
    5  - bits 1 e 5
    6  - bits 1 e 6
    7  - bits 1 e 7
    8  - bits 2 e 4
    9  - bits 2 e 5
    10 - bits 2 e 6
    11 - bits 2 e 7
    12 - bits 3 e 4
    13 - bits 3 e 5
    14 - bits 3 e 6
    15 - bits 3 e 7

** Seqncia de estados da simulao do pressionamento da tecla:

    LIVRE - BOUNCE_ENTRANDO - PRESSIONADA - BOUNCE_SAINDO - LIVRE

    OBS: 1. A varivel nciclos_pressionamento corresponde ao no. de ciclos desde
      o estado BOUNCE_ENTRANDO (ncio) at o estado BOUNCE_SAINDO (trmino).
         2. A varivel nciclos_bounce corresponde ao nmero de ciclos dos estados
      BOUNCE_ENTRANDO e BOUNCE_SAINDO (cada um deles).
         3. A varivel nciclos_estado_bounce corresponde ao no. de ciclos de
      durao de meio ciclo do bounce (considerando "onda quadrada") - ver
      abaixo.


** Explicaes sobre a simulao do bounce:

  O bounce  simulado de forma a produzir, no bit adequado da porta P1, uma onda
  quadrada, cujo semiciclo tem a durao, em ciclos de relgio, especificada em
  nciclos_estado_bounce. Para tanto considera-se, durante o bounce, que a tecla
  esteja ora em curto-circuito, ora em circuito aberto e recalcula-se o valor de
  leitura da porta P1 efetuando-se o AND entre os bits conectados  tecla. Como
  mais de uma tecla pode estar pressionada ao mesmo tempo faz-se necessrio
  repetir o clculo do valor de P1 at que o resultado permanea constante (ver
  algoritmo em computaValorP1() ). Note-se que se os bits conectados  tecla
  tiverem ambos o mesmo valor, durante o bounce, os mesmos no sero alterados.
*/

public class Teclado {
  //
  // CONSTANTES REFERENTES AOS ESTADOS DAS TECLAS
  //
  public static final int TECLA_LIVRE = 0;           // tecla no pressionada

  public static final int TECLA_PRESSIONADA = 1;     // tecla pressionada e
                                                     // estvel (s/ bounce)

  public static final int TECLA_BOUNCE_ENTRANDO = 2; // tecla recm pressionada
                                                     // efetuando bounce

  public static final int TECLA_BOUNCE_SAINDO = 3;   // tecla recm liberada
                                                     // efetuando bounce

  //
  //  VALORES DE CONFIGURAO DEFAULT
  //

  public static final int DEF_nciclos_pressionamento = 1700000;
  public static final int DEF_nciclos_bounce = 250000;
  public static final int DEF_nciclos_estado_bounce = 25000;

  //
  // VARIVEIS
  //

  public boolean ativo;               // indica teclado ativo ou no - usado
                                      // para controle externo.
  public int nciclos_pressionamento;  // no. de ciclos em que a tecla permanece
                                      // pressionada - inclui bounce

  public int nciclos_bounce;          // no. de ciclos de durao do bounce

  public int nciclos_estado_bounce;   // no. de ciclos de durao de durao de
                                      // meio ciclo do bounce (considerar onda
                                      // quadrada).

  public int nciclos_estavel;        // no. de ciclos em que a tecla
                                      // pressionada fica estvel (s/ bounce)

  private int nteclas_press;          // no. de teclas pressionadas

  private int[] estado_tecla = new int[16];  // estado atual da tecla(ver acima)

  private int[] cont_bounce = new int[16];   // contador de estados de bounce

  private int[] nciclos_estado_tecla = new int[16]; // no. de ciclos decorridos
                                                    // no estado atual da tecla.

  private boolean[] curto = new boolean[16]; // estado do bounce (curto circuito
                                             // ou circuito aberto).

  private Frame janela = null;     // tela grfica do teclado.
  private TelaTeclado tela = null;

  //
  // CRIA TECLADO
  //
  public Teclado() {
    ativo = true;

    nciclos_pressionamento = DEF_nciclos_pressionamento;
    nciclos_bounce = DEF_nciclos_bounce;
    nciclos_estado_bounce = DEF_nciclos_estado_bounce;

    reset();
  }

  //
  //  REINICIALIZA VARIVEIS DE ESTADO
  //
  public void reset() {
    for(int i=0; i<16; i++) {
      nteclas_press = 0;
      estado_tecla[i] = TECLA_LIVRE;
      cont_bounce[i] = -1;
      curto[i] = false;
      alteraCorBotao(i, 0);
    }
  }

  public void criaTela() {
    tela = new TelaTeclado();
    tela.controle = this;
    tela.isStandalone = true;
    janela = new Frame();
    janela.setTitle("TECLADO");
    janela.add(tela, BorderLayout.CENTER);
    tela.init();
    tela.start();
    janela.pack();
    Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
    janela.setLocation((d.width - janela.getSize().width) / 2, (d.height - janela.getSize().height) / 2);
    janela.setVisible(true);
  }

  public void mostraTela() {
    if(janela != null)
      janela.setVisible(true);
  }

  public void alteraCorBotao(int numbotao, int numcor) {
    if(janela != null)
      tela.botoes[numbotao].setForeground(tela.cores[numcor]);
  }

  //
  // FAZ O "TEMPO PASSAR" PARA A SIMULAO DO TECLADO
  //
  public void computaCiclos(int nciclos) {
    // se no h teclas pressionadas, retorna.
    if(nteclas_press == 0)
      return;

    // incrementa no. de ciclos do estado atual das teclas pressionadas e/ou
    // em estado de bounce e determina prx. estado.
    for(int t=0; t<16; t++) {
      // se a tecla est efetuando bounce ...
      if(estado_tecla[t] == TECLA_BOUNCE_ENTRANDO ||
        estado_tecla[t] == TECLA_BOUNCE_SAINDO)
      {
        // adiciona ciclos ao no. de ciclos do estado atual
        nciclos_estado_tecla[t] += nciclos;
        // se no. de ciclos superou no. ciclos de estado do bounce ...
        if(nciclos_estado_tecla[t] >= nciclos_estado_bounce) {
          // decrementa contador de estados do bounce
          cont_bounce[t]--;
          // se bounce terminou ...
          if(cont_bounce[t] < 0) {
            // se tecla recm pressionada - tecla permanece em curto
            if(estado_tecla[t] == TECLA_BOUNCE_ENTRANDO) {
              estado_tecla[t] = TECLA_PRESSIONADA;
              alteraCorBotao(t, 2);
              curto[t] = true;
            }
            // se tecla recm liberada - tecla permanece em aberto
            else {
//              System.out.println("Teclado: soltei tecla " + t);
              nteclas_press--;
              estado_tecla[t] = TECLA_LIVRE;
              alteraCorBotao(t, 0);
              curto[t] = false;
            }
          }
          // se bounce no terminou ...
          else
            // inverte estado do bounce : curto-circuito <-> circuito-aberto
            curto[t] = !curto[t];
          // zera contador de ciclos do estado (novo estado iniciou-se)
          nciclos_estado_tecla[t] = 0;
        }
      }
      // se a tecla est pressionada ...
      else if(estado_tecla[t] == TECLA_PRESSIONADA) {
        // adiciona ciclos ao no. de ciclos do estado atual
        nciclos_estado_tecla[t] += nciclos;
        // se no. de ciclos superou no. ciclos de estabilidade ...
        if(nciclos_estado_tecla[t] >= nciclos_estavel) {
          // inicia bounce "de sada"
          estado_tecla[t] = TECLA_BOUNCE_SAINDO;
          alteraCorBotao(t,1);
          // inicializa contador do bounce
          cont_bounce[t] = nciclos_bounce / nciclos_estado_bounce;
          // zera contador de ciclos do estado (novo estado iniciou-se)
          nciclos_estado_tecla[t] = 0;
        }
      }
    }
  }

  //
  // CALCULA E ATUALIZA O VALOR A SER LIDO NA PORTA P1 COM BASE NO ESTADO DE
  // PRESSIONAMENTO DAS TECLAS (CURTO-CIRCUITO OU CIRCUITO ABERTO)
  //
  public int computaValorP1() {
    int P1 = Memoria.memint[Memoria.P1];

    // se no h teclas pressionadas, retorna o valor do latch de P1.
    if(nteclas_press == 0)
      return P1;

    boolean[] bit_P1 = new boolean[8];
    boolean[] bit_P1_ant = new boolean[8];
    int i;

    // decompe valor do latch de P1 em bits
    for(i=0; i<8; i++)
      bit_P1[i] = (P1 & (1 << i)) > 0;

    boolean repetir;
    do {
      // copia valor atual de P1 (bit a bit)
      for(i=0; i<8; i++)
        bit_P1_ant[i] = bit_P1[i];

      // varre teclas recalculando os bits
      if(curto[ 0])
        bit_P1[0] = bit_P1[4] = bit_P1[0] && bit_P1[4];
      if(curto[ 1])
        bit_P1[0] = bit_P1[5] = bit_P1[0] && bit_P1[5];
      if(curto[ 2])
        bit_P1[0] = bit_P1[6] = bit_P1[0] && bit_P1[6];
      if(curto[ 3])
        bit_P1[0] = bit_P1[7] = bit_P1[0] && bit_P1[7];
      if(curto[ 4])
        bit_P1[1] = bit_P1[4] = bit_P1[1] && bit_P1[4];
      if(curto[ 5])
        bit_P1[1] = bit_P1[5] = bit_P1[1] && bit_P1[5];
      if(curto[ 6])
        bit_P1[1] = bit_P1[6] = bit_P1[1] && bit_P1[6];
      if(curto[ 7])
        bit_P1[1] = bit_P1[7] = bit_P1[1] && bit_P1[7];
      if(curto[ 8])
        bit_P1[2] = bit_P1[4] = bit_P1[2] && bit_P1[4];
      if(curto[ 9])
        bit_P1[2] = bit_P1[5] = bit_P1[2] && bit_P1[5];
      if(curto[10])
        bit_P1[2] = bit_P1[6] = bit_P1[2] && bit_P1[6];
      if(curto[11])
        bit_P1[2] = bit_P1[7] = bit_P1[2] && bit_P1[7];
      if(curto[12])
        bit_P1[3] = bit_P1[4] = bit_P1[3] && bit_P1[4];
      if(curto[13])
        bit_P1[3] = bit_P1[5] = bit_P1[3] && bit_P1[5];
      if(curto[14])
        bit_P1[3] = bit_P1[6] = bit_P1[3] && bit_P1[6];
      if(curto[15])
        bit_P1[3] = bit_P1[7] = bit_P1[3] && bit_P1[7];

      // se houve alterao nos bits, repete clculo
      repetir = false;
      if(nteclas_press > 1) {
        i = 0;
        while(i < 8 && !repetir) {
          if(bit_P1[i] != bit_P1_ant[i])
            repetir = true;
          i++;
        }
      }
    } while(repetir);

    // restaura valor de leitura P1 e atualiza
    P1 = 0;
    for(i=0; i<8; i++) {
      if(bit_P1[i])
        P1 += 1 << i;
    }
    return P1;
  }

  //
  // SIMULA O PRESSIONAMENTO DE UMA TECLA, ALTERANDO SEU ESTADO.
  //
  public void pressionaTecla(int numtecla) {
    //System.out.println("Teclado(): Apertei tecla " + numtecla);
    // incrementa no. de teclas pressionadas
    if(estado_tecla[numtecla] == TECLA_LIVRE)
      nteclas_press++;

    // inicia bounce "de entrada"
    estado_tecla[numtecla] = TECLA_BOUNCE_ENTRANDO;
    // inicializa contador de estados do bounce
    cont_bounce[numtecla] = nciclos_bounce / nciclos_estado_bounce;
    // zera contador de ciclos do estado atual (novo estado iniciou-se)
    nciclos_estado_tecla[numtecla] = 0;
    // inicializa tecla em circuito aberto
    curto[numtecla] = false;
    alteraCorBotao(numtecla, 1);

    // recalcula nciclos_estavel para o caso de haver tido alteraes
    // em nciclos_pressionamento e/ou nciclos_bounce
    nciclos_estavel = nciclos_pressionamento - 2*nciclos_bounce;
  }

  //
  // INFORMA O ESTADO ATUAL DA TECLA
  //
  public int estadoTecla(int numtecla) {
    return estado_tecla[numtecla];
  }

  public void atualizaTela() {
  }
}
